/* int gedits(row, column, string, string_size, mask)

   Get EDITed String from user console. 

   Line is kept full of blanks or characters for proper display.
   On exit, any trailing blanks are deleted.

*/ 
/* Parameter for debug
#define DEBUG
*/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "globals.h"

extern int  gtkey();
extern void setcur();
extern void wchar();
extern int  dos();
extern int give_error(char *msg);
extern void display_status();

static int srow,scol,fcol,ccol,insert_mode;
static char *end, *fcp, *lcp, *mend, *mfcp, *mlcp, *mstart, *start;

static char _Digits[]=" -.0123456789";
static char printed_value[32];

/* left -------------------------------------------------------
   Shift all characters left by one */

static void left (to)
char *to;

{   char *mp, *show;
    int offset;
    
    mp = to - start + mstart + 1;
    show = to;
    setcur(srow,ccol);
    for (offset = 0 ; to + 1 <= lcp; ++to,++mp)
    {   if (*mp == '_')
        {   if (*(mp-1) == '_' || offset != 0)
            {   *(to + offset) = *(to + 1);
                offset = 0;
            }
        }
        else
            --offset;
    }
    *to = '_';
    wstr(show,1);
    setcur(srow,ccol);
}

/* right -------------------------------------------------------
   Shift all characters right by one */

static void right (from)
char *from;

{   char *mp, *stop;
    int offset;

    for (offset = 1, stop = from, from = lcp - 1, mp = mlcp - 1;
         from >= stop; --from, --mp)
    {   /* Are both source and destination available? */
        if (*mp == '_' && *(mp + offset) == '_')
        {   *(from + offset) = *from;  /* Move character */ 
            offset = 1;                /* and cancel excess offset. */
        }
        else
            ++offset;  /* Can not move, increase offset. */
    }
    setcur(srow,ccol);
    wstr(stop,0);
}

/* show_line ----------------------------------------------------*/
/* Move buffer to screen and space fill empty end of line */

void show_line()
{
    char *line,*mask;

    setcur(srow,scol);
    for (line = start, mask = mstart ; *mask != '\0'; ++mask,++line)
        if (*mask == '_')
            wchar(*line);
        else
        {   wchar(*mask); *line=*mask;
        }
    setcur(srow,fcol);
}

/* ---------------------------------------------------------------- */
/* Identifies punctuation symbols in entry field */
int pnct(c)
char c;
{   return (c=='/' || c=='.' || c=='-' || c==':' || c== ',' ||
            c=='(' || c==')' || c==']');
}

/* clear_end -------------------------------------------------------
   Clears trailing blanks from end of line. */

void clear_end(st,bs,mk)
char *st, *mk;
int  bs;
{   char c;
    int ix;
    for(ix=bs-1; ix >= 0; --ix)
    {   if (pnct(*(mk+ix))) continue;
        if ((c=*(st+ix)) != '_' && c != ' ') break;
        *(st+ix)='_';
    }
    return;
}

static int valid_date(char *date)
{
    static int days[]={31,28,31,30,31,30,31,31,30,31,30,31};
    int day=0, month=0, year=0, year_chars=0, leap, which_char=0, valid, a_digit;
    char *p;

	p = date;
	while (*p !=0)
	  { if (which_char < 4 && (*p == '_' || *p == ' '))
			/* Underlines, spaces act as zeros, except in Year. */
		*p = '0';
	    a_digit = *p++ - '0';
	    if (a_digit >= 0 && a_digit <= 9)
		switch (which_char++)
		   { case 0:	/* Month */
		     case 1:
			month = month * 10 + a_digit;
			break;
		     case 2:	/* Day */
		     case 3:
			day = day * 10 + a_digit;
			break;
		     case 4:	/* Year */
		     case 5:
			year = year * 10 + a_digit;
			year_chars++;
			break;
		   }
	  }
	leap = (month == 2 && (year % 4 ==0));
	valid = (day == 0 && month == 0 && year_chars == 0); /* No date at all */
	if (valid)
	  for (p = date; *p != '\0'; p++)
	    if (*p == '0') *p = '_';	/* Convert zeros to underlines */

	valid = (valid ||
		 (month > 0 && month <= 12 &&
		  day > 0 && (day <= days[month-1]+leap) && year_chars == 2));
	return(valid);
}

/* gedits -------------------------------------------------------
    Gets a string with editing features */

int gedits(row,col,st,mask,afield)
    char *st,*mask;
    struct field *afield;
    int row,col;
{
    register int  c;        /* Current character */
    char save[80], *cp, *mp;
    int bufsize, fp=1, ucase=1;	/* if NULL is passed, do Text Uppercase */

    bufsize = strlen(mask);
    if (afield != NULL)
      {	fp = afield -> type;
	ucase = afield -> ucaseonly;
      }
    srow = row;
    fcol = scol = ccol = col;
    fcp = start = st;
    end = start + bufsize;
    mstart = mask;
    mend = mstart + bufsize;
    /* Underline fill to end of field for correct display. */
    *end = '\0';
    for (cp = st; (cp < end) && *cp != '\0'; ++cp) ;
    for ( ; cp < end ; ++cp) *cp = '_';
    for (mfcp = mstart; *mfcp != '_'; ++mfcp,++fcp,++fcol) ;
    mp = mfcp;
    cp = fcp;
    ccol = fcol;
    wchar('\1');
    show_line();
    /* Exit after line display, if not updating line */
    if(get_mode==0)
    {   wchar('\2');
        return(0);
    }
    /* Save this field for possible restore */
    memcpy(save,start,(unsigned)bufsize);
    cursize(insert_mode);
    /* Set left and right limits. */
    for (lcp = end, mlcp = mend ; *mlcp != '_'; --lcp,--mlcp) ;
    /* Input the line */
    stat_line = blanks;		/* Nothing on status line */
    while ((c=gtkey()) != EOF)
    { if (stat_line != blanks)	/* Erase previous status */
	{ stat_line = blanks;
	  display_status();
	  wchar('\1');		/* Inverse video */
	  setcur(srow,ccol);
	}
      switch(c) /* = c == ' ' ? '_' : c ) */
        {   case LEFT:		/* Left arrow */
	        while ( cp > fcp )
	        {   if (*(mp-1) == '_')
	            {   --ccol; --cp; --mp;
	                setcur(srow,ccol);
	                break;
	            }
	            --ccol; --cp; --mp;
		}
		break;
	    case RIGHT:		/* Right arrow */
		while ( cp < lcp )
		{   if (*(mp+1) == '_')
		    {   ++ccol; ++cp; ++mp;
			setcur(srow,ccol);
			break;
		    }
		    ++cp; ++ccol; ++mp;
		}
		break;
	    case INS:		/* Insert key */
		insert_mode = !insert_mode;
		cursize(insert_mode);
		break;
	    case BS:		/* Backspace */
		while ( cp > fcp )
		{   if (*(mp-1) == '_')
		    {   --ccol; --cp; --mp;
			setcur(srow,ccol);
			break;
		    }
		    --ccol; --cp; --mp;
		}
		/* Now delete the current character */
	    /* Delete key */
	    case DEL:
		/* Is there anything to delete ? */
		if (cp <= lcp) left(cp);
		break;
	    case HOME:		/* Home key */
		cp = fcp;
		mp = mfcp;
		setcur(srow,ccol=fcol);
		break;
	    case END:		/* End key */
		/* If not at end of data, go to end of data */
		for (cp = end, mp = mend ;
		      cp > fcp && *(cp-1) == *(mp-1) ; --cp, --mp) ;
		for ( ; cp <= lcp && *mp != '_' ; ++cp, ++mp) ;
		setcur(srow,ccol = cp - start + scol);
		break;
	    case CAN:		/* Cancel key - erase buffer and line */
		memcpy(start,mstart,(unsigned)bufsize);
		show_line();
		cp = fcp;
		mp = mfcp;
		break;
	    case ESC:		/* Escape key - restore initial input text */
		memcpy(start,save,(unsigned)bufsize);
		show_line();
	    /* These all exit retaining the current line */
	    case SH_TAB:
	    case CTL_LEFT:
	    case CTL_RIGHT:
	    case TAB:
	    case UP:
	    case DOWN:
	    case PGUP:
	    case PGDN:
	    case F1:
	    case F2:
	    case F3:
	    case F4:
	    case F5:
	    case F6:
	    case F7:
	    case F8:
	    case F9:
	    case F10:
	    case '\r':
	    case '\n':
		if (fp == 6 && !valid_date(st))	/* Date field with invalid date */
		  { give_error("\7Invalid Date");
		    break;
		  }
		cursize(0);
		clear_end(st,bufsize,mask);
		show_line();
		wchar('\2');
		return(c);
	    default:		/* If space available, insert the current character */
		if (ucase)	/* uppercase */
			c = toupper(c);
		if (fp == 2 || fp == 6)	/* NUMERIC or DATE field */
		  switch (c)
		     {	case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				break;
			case 32:
				c = '_';	/* Spaces go to underlines */
				break;
			default:
				c = 0;	/* disallow anything but those listed above */
		     }
		if (isprint(c))
		{   if (cp < lcp && insert_mode) right (cp);
		    if (cp <= lcp)
		    {   wchar (*cp++ = c); 
			++mp;
			/* Advance cursor to next valid entry column */
			for  (; cp <= lcp ; ++cp, ++mp)
			    if (*(mp) == '_') break;
			setcur(srow,ccol = cp - start + scol);
		    }
		    else wchar(BEL);	/* at end of field */
		}
		else wchar(BEL);	/* invalid char */
		break;
	}
	if (stat_line != blanks)
	    {	display_status();
		wchar('\1');
		setcur(srow,ccol);
	    }
    }
    cursize(0);
    wchar('\2');
    clear_end(st,bufsize,mask);
    return(c);
}

void showreal(row, col, afield)
    int row,col;
    struct field *afield;
{
    char *pv;		/* pointer into printed_value */
    char *dot;		/* pointer (near pv) where decimal point is */
    int  doti;		/* index to dot */
    char mask_over_print[32];	/* where we assemble what we print */
    char *mp;		/* pointer into mask_over_print */
    int  pvi, mpi;	/* index into printed_value, mask_over_print */
    int  ulcount;
    int  sigdigs;	/* Significant digits after decimal point */
    double val;

    if (afield -> type < 3)	/* Not a number field */
	return;

    val = afield -> value;
	/* Print spaces in field only if it's an entered (not computed)
	   field, and there's nothing in it, and its value is 0. */
    if (val != 0 || strcmp(afield->mask,afield->work) ||
	afield -> formula != NULL)
	{ sprintf(printed_value,"%26.10f\0",val); } /* 15 digits before, 10 after decimal point */
    else
	{ memset(printed_value,' ',26); printed_value[26]=0; }

    strcpy(mask_over_print,afield -> mask);
    if ((dot = strchr(mask_over_print,'.')) == NULL)
      { pv = printed_value+strlen(printed_value)-1;
	sigdigs = (*pv == ' ')?0:10;	/* No digits if blank, else start w/ 10 */
	while (*pv-- == '0') sigdigs--;
		/* sigdigs is # of significant digits after decimal point;
		   pvi is number of sig. digs. before dec. pt. */
	pvi = 15;
	pv = printed_value;
	while (*pv++ == ' ') pvi--;

	/* Find out how many underlines in the field */
	mp = mask_over_print; ulcount=0;
	while (*mp) if (*mp++ == '_') ulcount++;

	/* Will the integral & fractional digits, and a period, fit? */
	if (sigdigs + pvi + 1 > ulcount)
	  { sigdigs = ulcount - pvi - 1;	/* No -- truncate */
	  }

	ulcount = 0;	/* Now count from end of field backwards */
	doti = strlen(mask_over_print)-1;
	dot = mask_over_print + doti;
	while(doti > 0 && ulcount < sigdigs)
	  { if (*dot == '_')
		ulcount++;
	    dot--; doti--;
	  }
	if (doti < 0 || sigdigs <= 0)
	  { doti = strlen(mask_over_print);
	    dot = mask_over_print+doti;
	    if (*(dot-1) != '_')
		{ doti--; dot--; }	/* Trailing terminating char. in field */
	  }
	else
	  *dot = '.';
      }
    else	/* Use the found dot */
      {	doti = dot - mask_over_print;
      }

    mpi = doti-1;		pv = strchr(printed_value,'.')-1;
    mp = mask_over_print+mpi;	pvi= pv - printed_value;

	/* Copy digits to left of decimal */

    while (mpi >= 0)	   
			   /* while characters in mask */
      {	if (*mp == '_' && pvi >=0 )
			   /* Underline in mask = print here, */
			   /* but don't copy if no more numbers left */
          { *mp-- = *pv--; /* Copy from printed_value to mask_over_print */
	    pvi--;	   /* Back up printed_value index */
	  }
	mpi--;		/* go to previous character in mask_over_print & mask */
      }

	/* Copy digits to right of decimal */

    mpi = doti+1;		pv = strchr(printed_value,'.')+1;
    mp = mask_over_print+mpi;	pvi= pv - printed_value;

    while (*mp != '\0')		/* while place to print chars */
      {	if (*mp == '_' && *pv)	/* Copy into underlined areas */
	  *mp = *pv++;
	mp++;
      }

    mp = mask_over_print + strcspn(mask_over_print,_Digits);
    val = atof(mp);

    /* Translate spaces to underlines */
    strcpy(printed_value,mask_over_print);
    for (pv=printed_value; *pv != '\0'; pv++)
      if (*pv == 32) *pv = '_';

    if (row >= 0)
      {	setcur(row,col);
	wstr(mask_over_print,0);
      }
    for (pv=printed_value, mp=afield -> work; *pv && *mp ;)
	*mp++ = *pv++;
	
}

void get_value_from_field(struct field *afield)
{  char *p;
  double round;

	p = afield -> work;
	afield -> value = atof(p + strcspn(p,_Digits));

	round = (afield -> value > 0)?1:-1;
	switch (afield -> type)
	  { case 3:	/* Integers, truncate */
		afield -> value = (double)((long)(afield->value+.5*round));
		break;
	    case 4:	/* Dollars, round to cents */
		afield -> value = (double)((long)((afield->value+.005*round)*100))/100;
		break;
	  }
}

void update_value(struct field *afield)
{	showreal(-1,-1,afield);
}

/*
	This routine shifts everything to the left of the decimal
	point (if any) by one character and inserts a new character.
*/
static int shift_left_and_insert(struct field *afield, char to_insert)
{
    char *p, *q;

	p = afield -> work;
	q = afield -> mask;
	/* Now find first empty place to shift chars into */
	while (*p && *q && *q != '_')
	  { p++; q++; }
	if (*p != '_')
	  { wchar(BEL); return(0); }	/* No empty places */
	while (*p && *q == '_' && *(q+1)=='_')
	  { *p = *(p+1);
	    p++; q++;
	   }
	/* p winds up on the last blank before the end or decimal */
	*p = to_insert;
	get_value_from_field(afield);
/*	if (isdigit(to_insert))
		update_value(afield); */
	return(1);
}

/* getdlrs -------------------------------------------------------*/

int getdlrs(row, col, st, mask,  afield)
    int row,col;
    char *st, *mask;	/* st is current contents of field */
    struct field *afield;
{
    char initval[20];
    int c;	/* Current character */
    int cursor_start;
    int fraction = 0;
    int floating = 0;	/* TRUE if there's no decimal point in the mask */
    int i;
    char *work;
    char *p, *q;

    if (get_mode == 3) return(0);

    cursor_start = strcspn(mask,".");
    floating = mask[cursor_start] != '.';

    if (floating)
	cursor_start--;	/* Cursor on last char in field please */

    work = afield -> work;
    if (strcmp(work, mask))
      { p = work + strcspn(work,".");
	if (*p++ == '.')	/* If there's a value */
	  do fraction++;
	    while (isdigit(*p++));
      }

    strcpy(initval,afield->work);
    showreal(row, col, afield);

	/*  Input the value */
    while (1)
    {	get_value_from_field(afield);
	
	setcur(row,col);
	wstr(work,1);

#if 0
	setcur(24,0);
	printf("Mask: %-10.10s Work: %-10.10s Value: %12.4f, f%1d, formula:%-10.10s",
		mask, work, afield -> value, fraction, afield -> formula);
#endif

	setcur(row,col+cursor_start+((!floating)?fraction:0));
        c=gtkey();
        switch(c)
        {
/* Backspace */
        case BS:
        case DEL:
		/* When we're editing a floating point number, we
		   delete the last character. */

	    if (afield -> value ==0 && fraction == 0)
	      {	strcpy(work,mask);
		break;
	      }
	    if (floating)
		i = cursor_start;
	    else
		i = cursor_start + fraction - 1;

	    p = work + i;
	    q = mask + i;
		
	    if (fraction == 0 || floating)
	      {	while (i > 0)
		  { if (*q-- == '_')
			*p = *(p-1);
		    p--; i--;
		  }
	        p = work;
		q = mask;

 		/* Blank out the first character */
		while (*q != '_') { p++; q++; }
		*p = '_';

		if (fraction > 0) fraction--;
	      }
	    else
	      {	if (*q == '_') *p = '0';
		fraction--;
	      }
	    break;
/* Cancel key - delete value */
        case CAN:
            strcpy(work, mask);
	    fraction=0;
            break;
/* Escape key - restore original value. */
        case ESC:
            strcpy(afield -> work, initval);
            break;
	/* These all exit retaining the current line */
	case SH_TAB:
	case CTL_LEFT:
	case CTL_RIGHT:
	case TAB:
	case UP:
	case DOWN:
	case PGUP:
	case PGDN:
	case F1:
	case F2:
	case F3:
	case F4:
	case F5:
	case F6:
	case F7:
	case F8:
	case F9:
	case F10:
	case '\r':
	case '\n':
	    showreal(row,col,afield);
	    setcur(row,col);
	    wstr(afield -> work,1);
            return(c);
/* If space available, insert the current character */
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		if (afield -> type == 4 && fraction == 3)
		  { wchar (BEL); break; }
		if (floating && fraction)
		  fraction += shift_left_and_insert(afield,c);
		else
		  if (fraction == 0)
		    shift_left_and_insert(afield,c);
		  else
		    { p = mask + fraction + cursor_start;
		      q = work + fraction + cursor_start;
		      if (*p == '_') *q = c;
		      if (*p)
			fraction++;
		    }
		break;
	case '-':
		showreal(-1,-1,afield);			/* Print value nicely */
		p = strchr(afield -> work,'-');		/* Find first dash */
		if (p == NULL)	/* Was blank, or previously positive */
		  { if (floating || afield -> value != 0)
			p = strrchr(afield -> work,'_');
			/* Find last underline */
		    else
			p = strchr(afield->work,'.')-1;
			/* Find underline before first dot */
		  }
		if (p != NULL)
		  *p = ((*p == '-')?'_':'-');
		else
		  wchar(BEL);
		break;
	case '.':
		if (afield -> type >= 4 && fraction ==0)
			/* For Dollars or Real fields only */
		  { if (floating)
		      fraction = shift_left_and_insert(afield,'.');
		    else
		      fraction=1;
		  }
		else
		  wchar(BEL);
		break;
	default:
		wchar(BEL);
		break;
        }
    }
}



#ifdef DEBUG

char mask[] =    "____(__)____--______";
char xline[81] = "0123 45 6789  012345";
int get_mode;
int nor_attr = 0x07;
int inv_attr = 0x70;
int cur_attr = 0x07;

main()

{
extern init_keys(void);
extern exit_keys(void);

    int xlen=21;
    int i, xrow,xcol,xret;

    init_keys();
    xret = 0;
    while (xret != F9)
    {
        cls();
        get_mode=0;
        setcur(10,0);
        printf("Mode 0 - The line now contains:");
        xrow=11;
        xcol=20;
        setcur(xrow,xcol-1);
        wchar('>');
        setcur(xrow,xcol+21);
        wchar('<');
        gedits(xrow,xcol,xline,20,mask);
        get_mode=1;
        setcur(13,0);
        printf("Mode 1 - Test line edit on next line:");
        xrow=14;
        xcol=20;
        setcur(xrow,xcol-1);
        wchar('>');
        setcur(xrow,xcol+21);
        wchar('<');
        xret=gedits(xrow,xcol,xline,20,mask);
        setcur(15,0);
        printf("Return=%d\n",xret);   
        for (i=0; i<xlen; ++i) printf("%X ",xline[i]);
        getchar();
    }
    exit_keys();
}
#endif

