/* newform - Part of Form Filler program
*/

/* #define DEBUG */
#include <string.h>
#include <ctype.h>
#include "globals.h"
#define COMMENT ';'
#define SPACE ' '

extern int setscrn(char *);
extern int locscrn(char *,int,int);
extern int setcnst(char *);
extern int loccnst(char *);
extern int locctrl(char *);
extern fatal_line(char *the_bad_line,char *message,int err_no);
extern void display_status();
extern void window_size(int top_mgn,int bottom_mgn);

char err_hdr[]="~Header Line";
char err_hdr_eof[]="End of File in Header";
char err_ivl[]="~First Character (Valid: *=\\@^\\~; )";
static char SelectFDF[]="Select New Form Definition File";
static char *FDF = SelectFDF+11;

/* ---------------------------------------------------------------- */

/* Reads one line,skips over comments */
char *readone(char *ls, int reading_header)
{
char *rv, *lastchar;
    do rv = fgets(ls,132,form_file);
    while ((*ls == COMMENT || *ls == SPACE || *ls == '\n') && rv != NULL);
    if (rv == NULL)
      { if (reading_header)
	  fatal(err_hdr_eof,20);
      }
    else
      { lastchar = rv+strlen(rv)-1;
	if (*lastchar == '\n') *lastchar = '\0';
	if (debug) printf("%s\n",rv);
	if (reading_header && (*rv!='h' && *rv!='H') )
		fatal_line(ls,err_hdr,20);
      }
    return (rv);
}

/* ---------------------------------------------------------------- */
/* Identifies punctuation symbols in entry field */
int punct(c)
char c;
{

    return (c=='/' || c=='.' || c=='-' || c==':' || c==',' ||
	    c=='(' || c==')' || c==']' || c=='_' || c=='$' );
}

/* -------------------- keyword -------------------- */
/*
	Given a character pointer, returns a number indicating
	the keyword number (or -1 if none); updates that character pointer.
*/
int keyword(char *keywords[], char **input)
{  char *input_scan, *key_scan;
   int i, which_keyword, chars_matched;

	while (isspace(**input))
		(*input)++;		/* Skip over leading spaces & tabs */
	if (**input == '\0')
		return(-2);	/* Nothing there. */
	strupr(*input);
	which_keyword = -1;
	i = 0;
	key_scan = keywords[i];
	while ( which_keyword < 0 && *key_scan != '\0' )
	  { for (input_scan=*input, chars_matched = 0;
		*key_scan != '\0' && *key_scan == *input_scan;
		chars_matched++, key_scan++, input_scan++)
		;

	    /* See if we scanned up to space/tab/NL/null in input string,
	       being sure we matched at least 3 characters. */
	    if (*input_scan <= ' ' && (chars_matched >= 3 || !*key_scan))
		which_keyword = i;	/* found the keyword! */
	    else
		key_scan = keywords[++i];	/* go to next keyword */
	  }

	while (**input != '\0' && !isspace(**input))
	  (*input)++;			/* skip over keyword */

	while (isspace(**input))
	  (*input)++;			/* skip over the spaces & tabs */

	return(which_keyword);
}

/* -------------------- parse_options -------------------- */
int parse_options(char *options)
{ char *p;
  int an_option;
    strupr(options);

    screen_mode = quote_mode = ul_mode = data_mode = pause_mode =
    top_mgn = left_mgn = form_len = 0;
    form_feed = 1;
    for (p=options; *p != '\0'; )
      { an_option=keyword(header_keywords,&p);
	switch (an_option)
	  { case 0:	/* SCReen */
		screen_mode++;
		break;
	    case 1:	/* DATamode */
		data_mode++;
		break;
	    case 2:	/* NOFormfeed */
		form_feed = 0;
		break;
	    case 3:	/* UNDerline */
		ul_mode++;
		break;
	    case 4:	/* LEFtmargin */
		while (isdigit(*p))
		  left_mgn = left_mgn * 10 + *p++ - '0';
		break;
	    case 5:	/* TOPmargin */
		while (isdigit(*p))
		  top_mgn = top_mgn * 10 + *p++ - '0';
		break;
	    case 6:	/* QUOTES mode */
		quote_mode++;
		break;
	    case 7:	/* FORMLENGTH specified */
		while (isdigit(*p))
		  form_len = form_len * 10 + *p++ - '0';
		break;
	    case 8:	/* PAUSE option */
		pause_mode++;
		break;
	    case -1:	/* ??? Unknown keyword */
		return(0);
	  }		
      }
    if (screen_mode && data_mode)	/* Contradictory, assume DATA */
	screen_mode = 0;
    if (data_mode)			/* No top or left margins in DATA */
	left_mgn = top_mgn = 0;

    return(1);
}

static void memfree(char *memory)
{
	if (memory != NULL)
		free(memory);
}
/* ---------------------------------------------------------------- */
int newform()

{
int      tg;
unsigned origfree;	/* Original amount free */
char     c, ls[133], *lp, *cp;
struct field *afield, *nextfield;
FILE *tmp_file;

    strcpy(ls,temp1);
    tg=get_mode;
/* If file name is set in , try to open. Then, ask for name */
    do
    {   if (tg && get_mode!=3)
	switch(showfile(ls,SelectFDF,WFormDefn,1))
	      {
		case ESC:
		case F10: return(F10);
		case 0: break;
		case 1: return(1); /* Error (no files usually) */
	      }

	/* open the form definition file */
	tg=1;
	strupr(ls);

	if (strstr(ls,FormDefn) == NULL
	    && strstr(ls,".") == NULL) strcat(ls,FormDefn);
	if ((tmp_file = fopen(ls,"rt")) == NULL)
	{   give_error(not_found);
	    if (get_mode == 3) return(1);
	}
	else
	    stat_line=empty;
    } while (tmp_file == NULL);

    if (form_file_open)  /* Deallocate existing memory */
    {   afield = first_field;
	while (afield != NULL)
	    {	memfree(afield -> work);
		memfree(afield -> formula);
		memfree(afield -> mask);
		memfree(afield -> last);
		memfree(afield -> entry_line);
		nextfield = afield -> next;
		free(afield);
		afield = nextfield;
	    }
    }
	/* Remember that there are no fields in memory */
    first_field = last_field = cur_field = first_print_field = NULL;
    last_line = -1;	/* Last entry-screen line on the form */

    form_file = tmp_file;
    form_file_open = 1;

    if (debug)
    {   cls();
	printf("%s debug trace.\n\n",FDF);
	origfree = _memavl();
    }
    memcpy(form_name,ls,32);

    if (get_mode != 3)
      {	cls();		/* clear screen so errors will occur over clear area */
	setcur(0,0);
	printf("Loading %s %s\n\n",FDF,form_name);
      }

    /* Header processing */
    /* Line one contains form title */
    readone(ls,1);
    cp = &ls[1];
    while (isspace(*cp)) cp++;	/* Skip leading spaces */
    strncpy(form_title,cp,71);
    form_title[71]='\0';	/* Always null terminate */

    window_size(*form_title?1:0,0);	/* Title, if any, uses top line. */

    /* Line two contains optional output file name */
    readone(ls,1);
    if (strlen(temp2) < 1)
      {	cp = &ls[1];
	while (isspace(*cp)) cp++;
	sscanf(cp,"%s",temp2);
	strcpy(temp2,Printer); /* Default output name */
      }
    strupr(temp2);

    /* Line three contains options */
    readone(ls,1);

    /* Parse options.  If any problems, abort. */
    if (!parse_options(&ls[1]))
	fatal_line(ls,err_hdr,20);
    cur_top=0;

/* Line processing for remainder of file */
    while ((lp = readone(ls,0)) != NULL)
    {   c=*lp;
	if (c == PADCHAR) break;  /* For Wordstar & others */
	switch(c)
	  { case SCREEN_LINE:
		setscrn(lp);
		break;
	    case SCREEN_LOC:
		locscrn(lp,-1,-1);
		break;
	    case CNST_LINE:
		setcnst(lp);
		break;
	    case CNST_LOC:
		loccnst(lp);
		break;
	    case CNST_CTRL:
		locctrl(lp);
		break;
	    default:
		fatal_line(ls,err_ivl,21);
	 }
    }
    if (fclose(form_file) == EOF)
	fatal("FDF file close.",3);

#ifdef DEBUG
    if(debug)
	show_field_list();
#endif

    if(debug)
      {	printf("\n\n%u bytes used, %u free.  %s\n",
		origfree-_memavl(),_memavl(),any_key);
	gtkey();
      }

    for (first_data_field = first_field;
	 first_data_field -> type <=0;	/* Keep looping, skipping non-data nodes */
	 first_data_field = first_data_field -> next)
	 ;

    return(newout());
}
