/*
shows the *.FDF files in the current directory
 */

#include "globals.h"
#include "std.h"
#include "doslib.h"
#include <search.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>
#include <stdlib.h>

extern give_error(char *text);
extern void display_status();
extern void draw_a_box(int y1, int x1, int y2, int x2, int dflag);
extern void fatal(char *, int);

/* constants for DOS file-matching routines */
#define FILESPEC        13      /* maximum filespec + NUL */
#define MAXCOL          80

/* allocation quantities */
#define MAX_FILES 100
#define CALLOC(n,x) ((x *)calloc(n,sizeof(x)))

char Selection[] =
	"Space, Backspace, Arrows, & \1Enter\2 to Select  \1F10\2 Main Menu";

/* package data */
static  struct find_t buf;	/* private disk buffer */
static  char path[MAXPATH + 1]; /* working pathname */
static  int nfiles;		/* number of file names */
static  int ncols;		/* number of columns to output */
static  int nlines;		/* number of lines to output */
static  int mod;		/* mod columns in last row */
static  int maxlen;		/* screen columns per path name */
static  int x_off=8, y_off=5;


/* ---------------------------------------------------------------- *
   last_ch -- returns pointer to last character in string
*/
char last_ch(s)
char *s;
{
char *cp;
    for (cp=s; *cp != '\0'; ++cp);
    return (*(--cp));
}
/* end last_ch */

/* ---------------------------------------------------------------- *
   ls_multi -- list items in multiple columns that
   vary in width and number based on longest item size
*/

int ls_multi(buf, nelem)
char buf[MAX_FILES+1][FILESPEC];
int nelem;
{
    int i, j;
    int errcount = 0;
    int n;                  /* number of items in list */
    int len;
    /* get length of longest pathname and calculate number
       of columns and lines (col width = maxlen + 1)
    */
    n = 0;
    maxlen = 0;
    for (n = 0; n < nelem; ++n)
        if ((len = strlen(buf[n])) > maxlen) maxlen = len;
    maxlen += 2;	/* Allow 2 spaces between */

    nlines = (n < 15) ? n : 15;
    ncols = n / nlines + 1;
    mod = n % nlines;
    if (!mod)		/* Don't need extra column */
	{ ncols--;
	  mod = nlines;
	}

    draw_a_box(y_off-1,x_off-2,y_off+nlines,x_off+(maxlen)*ncols,2);
    setcur(y_off,x_off);

    /* output multi-column list */
    for (i = 0; i < nlines; )
    {   for (j = 0; j+i < n; j += nlines)
	  { len = maxlen - strlen(buf[i+j]);
	    wstr(strlwr(buf[i+j]),1);
	    while (len-- > 0) wchar(' ');
	  }
	 setcur(++i+y_off,x_off);
    }
    return (errcount);
} /* end ls_multi() */
/* ------------------------ find_x_y ----------------------------- *
*/
int find_x_y(ix, x, y)
int ix, *x,*y;
{
	*y = ix % nlines;
	*x = ix / nlines;
}
/* ------------------------ pick_one ----------------------------- *
*/
int pick_one(buf)
char buf[MAX_FILES+1][FILESPEC];

{
    int c, ix, ixr, x, y;

    /* Set index to first name */
    ix = x = y = 0;
    for (;;)
    {   /* highlight this selection */
        setcur(y+y_off,x*(maxlen)+x_off);
        wchar('\1'); wstr(buf[ix],0); wchar('\2');
        c = gtkey();
        setcur(y+y_off,x*(maxlen)+x_off);
        /* restore selection to normal */
        wchar('\2'); wstr(buf[ix],0);
        /* Look for special keys */
        switch (c)
        {   case '\n':
                return(ix);
            case ESC:
            case F10:
                return(F10); /* Cancel selection */
            case HOME:
                ix=x=y=0;
                break;
            case END:
                ix=nfiles;
            case UP:
            case BS:
                if (--ix < 0) ix = nfiles - 1;
                find_x_y(ix,&x,&y);
                break;
            case DOWN:
            case ' ':
                if (++ix >= nfiles) ix = 0;
                find_x_y(ix,&x,&y);
                break;
            case LEFT:
		if (--x < 0)
		  { if (--y < 0)
		      y = nlines-1;
		    x = ncols - 1 - (y >= mod);
			/* Go left 1 column if column not full on this line*/
		  }
                break;
            case RIGHT:
		if (++x > ncols - 1 - (y >= mod))
		  { x = 0;
		    if (++y >= nlines)
		      y = 0;
		  }
                break;
	    default:
		c = tolower(c);
		if (isalnum(c))
		  { ixr = ix; ix = 0;
		    while (ix < nfiles && buf[ix][0]<c) ix++;
		    if (ix >= nfiles)
		      ix = ixr;		/* No files by that letter */
		    find_x_y(ix,&x,&y);
		  }
        }
	ix = y + (nlines * x);
    }
}

void pick_file_prompt(char *message)
{   int ch, i;

	if (getcwd(path, MAXPATH) == NULL)
		fatal(sys_errlist[2],3);

	cls();
	setcur(1,x_off);
	wstr(message,1);
	setcur(2,x_off);
	wstr("Directory: ",1);
	wstr(path,1);
	i = strlen(message);
	ch= strlen(path)+12;
	draw_a_box(0,x_off-2,3,x_off+((i>ch)?i:ch),1);
}

/* ---------------------------------------------------------------- *
*/
int showfile(choice,message,filter,no_ext)
char *choice, *filter, *message;
int no_ext;
{
typedef char OUTA[FILESPEC];
    int i;
    char *pt;
    OUTA *bp;	/* output buffer pointer */

	pick_file_prompt(message);
	setcur(23,0);
	wstr(Selection,0);
	display_status();

	setcur(y_off,x_off);
	bp = CALLOC(MAX_FILES+1,OUTA);
	if (bp == NULL)
		fatal(err_mem,2);
	memset(bp,'\0',(MAX_FILES+1) * FILESPEC);
	/* use current directory */
	if (last_ch(path) != '\\') strcat(path,"\\");
	strcat(path,filter);			/* *.FIF or *.FDF */
	if(_dos_findfirst(path,_A_NORMAL,&buf))	/* Find normal files */
	{ wstr("\nNo files found.",1);
	  wstr(any_key,1);
	  *choice = '\0';
	  free(bp);
	  gtkey();
	  return(1);
	}
	/* add file or directory to the buffer */
	nfiles = 0;
        do
        {   if (buf.name[0] == '.') continue;
            strcpy(bp[nfiles],buf.name);
            if (no_ext && (pt = strstr(bp[nfiles],".")) != NULL)
                *pt = '\0';
            ++nfiles;
        } while (_dos_findnext(&buf) == 0 && nfiles <= MAX_FILES);
	if(nfiles > MAX_FILES)
	  { give_error("More than 100 files.");
	    display_status();
	    setcur(y_off,x_off);
	    nfiles = MAX_FILES;
	  }
        /* sort and list */
	qsort(bp, nfiles, FILESPEC, strcmp);
	ls_multi(bp, nfiles);
	if ((i = pick_one(bp)) == F10) return(F10);
	strcpy(choice,bp[i]);
	free(bp);
	return (0);
}
